@@ -43,6 +43,7 @@ class SQL(object):
         import os
         import re
         import sqlalchemy
+        import sqlalchemy.orm
         import sqlite3
 
         # Require that file already exist for SQLite
@@ -56,6 +57,7 @@ class SQL(object):
         # Create engine, disabling SQLAlchemy's own autocommit mode, raising exception if back end's module not installed
         self._engine = sqlalchemy.create_engine(url, **kwargs).execution_options(autocommit=False)
 
+        # Get logger
         self._logger = logging.getLogger("cs50")
 
         # Listener for connections
@@ -71,12 +73,11 @@ class SQL(object):
                 cursor.execute("PRAGMA foreign_keys=ON")
                 cursor.close()
 
-            # Autocommit by default
-            self._autocommit = True
-
         # Register listener
         sqlalchemy.event.listen(self._engine, "connect", connect)
 
+        # Autocommit by default
+        self._autocommit = True
 
         # Test database
         disabled = self._logger.disabled
@@ -96,9 +97,9 @@ class SQL(object):
 
     def _disconnect(self):
         """Close database connection."""
-        if hasattr(self, "_connection"):
-            self._connection.close()
-            delattr(self, "_connection")
+        if hasattr(self, "_session"):
+            self._session.remove()
+            delattr(self, "_session")
 
     @_enable_logging
     def execute(self, sql, *args, **kwargs):
@@ -113,7 +114,7 @@ class SQL(object):
         import warnings
 
         # Parse statement, stripping comments and then leading/trailing whitespace
-        statements = sqlparse.parse(sqlparse.format(sql, keyword_case="upper", strip_comments=True).strip())
+        statements = sqlparse.parse(sqlparse.format(sql, strip_comments=True).strip())
 
         # Allow only one statement at a time, since SQLite doesn't support multiple
         # https://docs.python.org/3/library/sqlite3.html#sqlite3.Cursor.execute
@@ -129,8 +130,9 @@ class SQL(object):
         # Infer command from (unflattened) statement
         for token in statements[0]:
             if token.ttype in [sqlparse.tokens.Keyword, sqlparse.tokens.Keyword.DDL, sqlparse.tokens.Keyword.DML]:
-                if token.value in ["BEGIN", "DELETE", "INSERT", "SELECT", "START", "UPDATE"]:
-                    command = token.value
+                token_value = token.value.upper()
+                if token_value in ["BEGIN", "DELETE", "INSERT", "SELECT", "START", "UPDATE"]:
+                    command = token_value
                     break
         else:
             command = None
@@ -275,33 +277,34 @@ class SQL(object):
             # Infer whether app is defined
             assert flask.current_app
 
-            # If new context
-            if not hasattr(flask.g, "_connection"):
+            # If no sessions for any databases yet
+            if not hasattr(flask.g, "_sessions"):
+                setattr(flask.g, "_sessions", {})
+            sessions = getattr(flask.g, "_sessions")
 
-                # Ready to connect
-                flask.g._connection = None
+            # If no session yet for this database
+            # https://flask.palletsprojects.com/en/1.1.x/appcontext/#storing-data
+            # https://stackoverflow.com/a/34010159
+            if self not in sessions:
 
-                # Disconnect later
-                @flask.current_app.teardown_appcontext
-                def shutdown_session(exception=None):
-                    if flask.g._connection:
-                        flask.g._connection.close()
+                # Connect to database
+                sessions[self] = sqlalchemy.orm.scoping.scoped_session(sqlalchemy.orm.sessionmaker(bind=self._engine))
 
-            # If no connection for context yet
-            if not flask.g._connection:
-                flask.g._connection = self._engine.connect()
+                # Remove session later
+                if _teardown_appcontext not in flask.current_app.teardown_appcontext_funcs:
+                    flask.current_app.teardown_appcontext(_teardown_appcontext)
 
-            # Use context's connection
-            connection = flask.g._connection
+            # Use this session
+            session = sessions[self]
 
         except (ModuleNotFoundError, AssertionError):
 
             # If no connection yet
-            if not hasattr(self, "_connection"):
-                self._connection = self._engine.connect()
+            if not hasattr(self, "_session"):
+                self._session  = sqlalchemy.orm.scoping.scoped_session(sqlalchemy.orm.sessionmaker(bind=self._engine))
 
-            # Use this connection
-            connection = self._connection
+            # Use this session
+            session = self._session
 
         # Catch SQLAlchemy warnings
         with warnings.catch_warnings():
@@ -321,10 +324,10 @@ class SQL(object):
 
                 # Execute statement
                 if self._autocommit:
-                    connection.execute(sqlalchemy.text("BEGIN"))
-                result = connection.execute(sqlalchemy.text(statement))
+                    session.execute(sqlalchemy.text("BEGIN"))
+                result = session.execute(sqlalchemy.text(statement))
                 if self._autocommit:
-                    connection.execute(sqlalchemy.text("COMMIT"))
+                    session.execute(sqlalchemy.text("COMMIT"))
 
                 # Check for end of transaction
                 if command in ["COMMIT", "ROLLBACK"]:
@@ -357,7 +360,7 @@ class SQL(object):
                 elif command == "INSERT":
                     if self._engine.url.get_backend_name() in ["postgres", "postgresql"]:
                         try:
-                            result = connection.execute("SELECT LASTVAL()")
+                            result = session.execute("SELECT LASTVAL()")
                             ret = result.first()[0]
                         except sqlalchemy.exc.OperationalError:  # If lastval is not yet defined in this session
                             ret = None
@@ -533,3 +536,10 @@ def _parse_placeholder(token):
 
     # Invalid
     raise RuntimeError("{}: invalid placeholder".format(token.value))
+
+
+def _teardown_appcontext(exception=None):
+    """Closes context's database connection, if any."""
+    import flask
+    for session in flask.g.pop("_sessions", {}).values():
+        session.remove()
